@@ -7,6 +7,7 @@
 import string
 import os.path
 
+from Action import find_primary_action
 from Dictlist import varname2dictlist, dictlist2str
 from RecPos import RecPos, rpcopy
 from Util import *
@@ -222,8 +223,6 @@ def add_buildrule(rpstack, work, recdict,
     if type != "totype":
         work.add_dictlist_nodes(targetlist)
 
-    br_rpstack = rpstack        # rpstack from the build rule
-
     # When there is no direct build action for the source or there
     # are multiple sources, compile each source into an object and
     # build all the objects into the target.
@@ -251,58 +250,50 @@ def add_buildrule(rpstack, work, recdict,
         add_onestep_build(rpstack, work, recdict, type, cmd_attr, targetlist,
                             build_attr, sourcelist)
     else:
-        objtype = cmd_attr.get("objecttype")
+        a = cmd_attr.get("objecttype")
+        if a:
+            objtypes = string.split(a, ',')
+        else:
+            objtypes = None
         buildaction = cmd_attr.get("buildaction")
         if type == "dll":
-            if not objtype:
-                objtype = "dllobject"
             if not buildaction:
                 buildaction = "builddll"
         elif type == "lib":
-            if not objtype:
-                objtype = "libobject"
             if not buildaction:
                 buildaction = "buildlib"
         elif type == "ltlib":
-            if not objtype:
-                objtype = "ltobject"
             if not buildaction:
                 buildaction = "buildltlib"
         elif type == "program":
-            if not objtype:
-                objtype = "object"
             if not buildaction:
                 buildaction = "build"
         elif type == "totype":
-            objtype = targetlist[0]["name"]
+            objtypes = [ targetlist[0]["name"] ]
         else:
-            if not objtype:
-                recipe_error(rpstack, _("Missing objecttype attribute"))
             if not buildaction:
                 recipe_error(rpstack, _("Missing buildaction attribute"))
             if installvar is None:
                 installvar = "INSTALL_EXEC"
 
+        if not objtypes:
+            # Obtain the supported object types from the buildaction.
+            act = find_primary_action(buildaction)
+            if not act:
+                if type not in ["dll", "lib", "ltlib", "program"]:
+                    recipe_error(rpstack, _("Missing objecttype attribute"))
+                else:
+                    recipe_error(rpstack, _("No primary %s action defined")
+                                                                 % buildaction)
+            objtypes = act.get_in_types()
+
         objectsuffix = cmd_attr.get("objectsuffix")
         if type == "totype" and targetlist[0].get("suffix"):
             objectsuffix = targetlist[0].get("suffix")
-        sufname = None
-        if not objectsuffix:
-            if objtype == "dllobject":
-                sufname = "DLLOBJSUF"
-            elif objtype == "libobject":
-                sufname = "LIBOBJSUF"
-            elif objtype == "ltobject":
-                sufname = "LTOBJSUF"
-            elif objtype == "object":
-                sufname = "OBJSUF"
-
         objectprefix = cmd_attr.get("objectprefix")
         if type == "totype" and targetlist[0].get("prefix"):
             objectprefix = targetlist[0].get("prefix")
 
-        found_cpp = 0
-
         # Go over all source files and figure out a way to turn each one
         # into an object file.
         build_obj = []
@@ -310,104 +301,29 @@ def add_buildrule(rpstack, work, recdict,
             node = work.get_node(si["name"], 1, si)
             in_ftype = node.get_ftype(recdict)
 
-            if in_ftype == objtype:
-                # The source is an object file: add it directly.
+            if in_ftype in objtypes:
+                # The source is supported by the build action: add it directly.
                 build_obj.append(si)
             else:
-                # Make the object name by prepending $BDIR and changing the
-                # extension to $OBJSUF/$LIBOBJSUF/$DLLOBJSUF/nothing.
-                n = srcitem2obj(recdict, si["name"], attrdict = si,
-                                                         sufname = sufname)
-                if objectprefix:
-                    n = os.path.join(os.path.dirname(n),
-                                            objectprefix + os.path.basename(n))
-                if objectsuffix:
-                    n = n + objectsuffix
-                bsi = {"name" : n}
-                build_obj.append(bsi)
+                # Loop over the types that the build action supports, turn
+                # the source in the first one that is possible.
+                names = ''
 
-                if in_ftype == "cpp":
-                    found_cpp = 1
+                objname = None
+                for objtype in objtypes:
+                    if objtype == "default":
+                        continue
+                    objname = add_build_one(rpstack, work, recdict, node, si, in_ftype, objtype, objectprefix, objectsuffix, build_attr, cleanfiles)
+                    if objname:
+                        break
+                    if names:
+                        names = names + '/'
+                    names = names + objtype
 
-                # If there is no build rule for the object file, add one.
-                target = work.find_node(n)
-                if target and target.get_first_build_dependency():
-                    # There already is a dependency for this target.
-                    if len(si) > 1:
-                        msg_warning(recdict, _('%s: Ignoring attributes for "%s", a build rule already exists') % (str(rpstack[-1]), si["name"]))
-                else:
+                if objname is None:
+                    recipe_error(rpstack, _('Do not know how to make an %s out of "%s"') % (names, si["name"]))
 
-                    # Find the route to the object file.
-                    route = work.find_route(in_ftype, objtype)
-                    if not route:
-                        recipe_error(br_rpstack, _('Do not know how to make an %s out of "%s"') % (objtype, si["name"]))
-
-                    # Inits to avoid a warning from pychecker.
-                    trg = None
-                    out_ftype = None
-
-                    # Loop over all steps in the route.
-                    for i in range(len(route.steplist)):
-                        # Get the action and filename by processing the line.
-                        # TODO: catch errors
-                        route.rpstack[-1].line_nr = route.lnumlist[i] - 1
-                        fp = ParsePos(route.rpstack, string = "_x = "
-                                                + route.steplist[i] + '\n')
-                        rd = get_build_recdict(recdict, route.recdict,
-                                     route.rpstack, keep_current_scope = 1,
-                                     xscope = build_attr.get("scope"))
-                        rd["source"] = si["name"]
-                        Process(fp, rd, 0)
-                        l = string.split(rd["_x"], None, 1)
-                        action = l[0]
-
-                        if i == 0:
-                            src_dict = si.copy()
-                            src = si["name"]
-                            dnode = node
-                        else:
-                            src_dict = {"name" : trg}
-                            src = trg
-                            in_ftype = out_ftype
-                            dnode = work.find_node(src)
-
-                        # Set "depdir" for the auto-depend recipe.  Required if the
-                        # source file is built twice with different $BDIR.
-                        # Also add the recipe to the files to be cleaned now,
-                        # because $BDIR may change afterwards.
-                        if not src_dict.get("depdir"):
-                            src_dict["depdir"] = get_var_val_int(recdict, "BDIR")
-                        depaction, recipe = find_autodep_items(work, recdict,
-                                                     in_ftype, dnode, src_dict)
-                        if depaction:
-                            add_cleanfiles(recdict, recipe.get_name())
-
-                        if i == len(route.steplist) - 1:
-                            trg = n
-                            out_ftype = objtype
-                        else:
-                            trg = l[1]
-                            out_ftype = route.typelist[i + 1][0]
-                            if (not os.path.isabs(trg)
-                                    and string.find(trg, "build-") != 0):
-                                trg = os.path.join(get_var_val_int(recdict,
-                                                              "BDIR"), trg)
-
-                        # Add a dependency to execute the action.
-                        cmd = "  :do %s {target = %s} %s" % (action, trg, src)
-                        msg_depend(recdict, _('Adding dependency:\n\t%s : %s\n\t%s')
-                                        % (trg, dictlist2str([src_dict]), cmd))
-                        src_dict["filetype"] = in_ftype
-                        trg_dict = {"name": trg, "filetype" : out_ftype}
-                        work.add_dependency(br_rpstack,
-                               Depend([ trg_dict ], build_attr,
-                                   [ src_dict ], work,
-                                   rpcopy(route.rpstack, route.lnumlist[i]),
-                                       cmd + '\n', recdict = recdict))
-                        cleanfiles.append({"name": trg})
-
-                        if in_ftype == "cpp":
-                            found_cpp = 1
+                build_obj.append({"name" : objname})
 
 
     if type != "totype":
@@ -418,15 +334,12 @@ def add_buildrule(rpstack, work, recdict,
             #       do buildaction {target = $target} $source
             build_obj_str = dictlist2str(build_obj)
             cmd = ("  :do %s {target = $+target} $source" % buildaction)
-            if found_cpp:
-                # Need to use $CXX for building when $LD isn't set.
-                cmd = "  USECXXLD = yes\n" + cmd
 
             msg_depend(recdict, _('Adding dependency:\n\t%s : %s\n\t%s')
                                            % (targets_str, build_obj_str, cmd))
-            work.add_dependency(br_rpstack,
+            work.add_dependency(rpstack,
                        Depend(targetlist, build_attr, build_obj, work,
-                                 br_rpstack, cmd + '\n', recdict = recdict))
+                                       rpstack, cmd + '\n', recdict = recdict))
 
         cleanfiles.extend(targetlist)
 
@@ -444,6 +357,110 @@ def add_buildrule(rpstack, work, recdict,
     add_distfiles(recdict, dictlist2str(l, Expand(0)))
 
 
+def add_build_one(rpstack, work, recdict, node, si, in_ftype,
+                  objtype, objectprefix, objectsuffix, build_attr, cleanfiles):
+    """
+    Add build rule to turn one source node into a file of type "objtype".
+    If successful return name of target file, None otherwise.
+    """
+    sufname = None
+    if not objectsuffix:
+        if objtype == "dllobject":
+            sufname = "DLLOBJSUF"
+        elif objtype == "libobject":
+            sufname = "LIBOBJSUF"
+        elif objtype == "ltobject":
+            sufname = "LTOBJSUF"
+        elif objtype == "object":
+            sufname = "OBJSUF"
+
+    # Make the object name by prepending $BDIR and changing the
+    # extension to $OBJSUF/$LIBOBJSUF/$DLLOBJSUF/nothing.
+    n = srcitem2obj(recdict, si["name"], attrdict = si, sufname = sufname)
+    if objectprefix:
+        n = os.path.join(os.path.dirname(n), objectprefix + os.path.basename(n))
+    if objectsuffix:
+        n = n + objectsuffix
+
+    # If there is no build rule for the object file, add one.
+    target = work.find_node(n)
+    if target and target.get_first_build_dependency():
+        # There already is a dependency for this target.
+        if len(si) > 1:
+            msg_warning(recdict, _('%s: Ignoring attributes for "%s", a build rule already exists') % (str(rpstack[-1]), si["name"]))
+    else:
+
+        # Find the route to the object file.
+        route = work.find_route(in_ftype, objtype, use_actions = 1)
+        if not route:
+            return None
+
+        # Add attributes from ":route" to the target node.
+        work.get_node(n, add = 1, dict = route.targetattr)
+
+        # Inits to avoid a warning from pychecker.
+        trg = None
+        out_ftype = None
+
+        # Loop over all steps in the route.
+        for i in range(len(route.steplist)):
+            # Get the action and filename by processing the line.
+            # TODO: catch errors
+            route.rpstack[-1].line_nr = route.lnumlist[i] - 1
+            fp = ParsePos(route.rpstack, string = "_x = "
+                                                    + route.steplist[i] + '\n')
+            rd = get_build_recdict(recdict, route.recdict,
+                                         route.rpstack, keep_current_scope = 1,
+                                              xscope = build_attr.get("scope"))
+            rd["source"] = si["name"]
+            Process(fp, rd, 0)
+            l = string.split(rd["_x"], None, 1)
+            action = l[0]
+
+            if i == 0:
+                src_dict = si.copy()
+                src = si["name"]
+                dnode = node
+            else:
+                src_dict = {"name" : trg}
+                src = trg
+                in_ftype = out_ftype
+                dnode = work.find_node(src)
+
+            # Set "depdir" for the auto-depend recipe.  Required if the source
+            # file is built twice with different $BDIR.  Also add the recipe to
+            # the files to be cleaned now, because $BDIR may change afterwards.
+            if not src_dict.get("depdir"):
+                src_dict["depdir"] = get_var_val_int(recdict, "BDIR")
+            depaction, recipe = find_autodep_items(work, recdict,
+                                                     in_ftype, dnode, src_dict)
+            if depaction:
+                add_cleanfiles(recdict, recipe.get_name())
+
+            if i == len(route.steplist) - 1:
+                trg = n
+                out_ftype = objtype
+            else:
+                trg = l[1]
+                out_ftype = route.typelist[i + 1][0]
+                if (not os.path.isabs(trg) and string.find(trg, "build-") != 0):
+                    trg = os.path.join(get_var_val_int(recdict, "BDIR"), trg)
+
+            # Add a dependency to execute the action.
+            cmd = "  :do %s {target = %s} %s" % (action, trg, src)
+            msg_depend(recdict, _('Adding dependency:\n\t%s : %s\n\t%s')
+                                        % (trg, dictlist2str([src_dict]), cmd))
+            src_dict["filetype"] = in_ftype
+            trg_dict = {"name": trg, "filetype" : out_ftype}
+            work.add_dependency(rpstack,
+                   Depend([ trg_dict ], build_attr, [ src_dict ], work,
+                                      rpcopy(route.rpstack, route.lnumlist[i]),
+                                                cmd + '\n', recdict = recdict))
+            cleanfiles.append({"name": trg})
+
+    return n
+
+
 def add_onestep_build(rpstack, work, recdict,
                            type, cmd_attr, targetlist, build_attr, sourcelist):
     """Adds an onestep build"""
