@@ -6,6 +6,7 @@
 
 import os
 import os.path
+import string
 
 import Global
 from Node import Node
@@ -16,7 +17,7 @@ from ParsePos import ParsePos
 from RecPos import rpcopy, RecPos
 from Util import *
 from Depend import Depend, depend_auto
-from Remote import url_time
+from Remote import url_time, is_url
 from VersCont import fetch_nodelist
 from Cache import local_name
 from Sign import get_old_sign, get_new_sign, sign_clear, sign_clear_target
@@ -36,23 +37,26 @@ def dictlist2shellstr(list, key):
        shell command.
        TODO: let this depend on the type of shell used.
        """
-    str = ''
+    res = ''
     for i in list:
-        if str:
-            str = str + ' '
+        if res:
+            res = res + ' '
         for c in i[key]:
             if string.find("'\" \t", c) >= 0:
-                str = str + "\\"
-            str = str + c
-    return str
+                res = res + "\\"
+            res = res + c
+    return res
 
 
-def exec_commands(dep, recdict, target, filetype = None):
+def exec_commands(dep, recdict, target, filetype = None, really_build = 1):
     """
     Execute the build commands "dep.commands" to build the Node "target", using
     the info in the Depend object "dep".
     "filetype" is the value for the "filetype" variable (used for automatic
     dependencies).
+    When "really_build" is non-zero we are building normally.  When
+    "really_build" is zero we are executing the ">always" and ">nobuild"
+    sections of a command block.
     """
     work = getwork(recdict)
 
@@ -69,6 +73,8 @@ def exec_commands(dep, recdict, target, filetype = None):
                                rpstack = dep.rpstack,
                                xscope = dep.get_scope_names(work))
 
+    new_recdict["_really_build"] = really_build
+
     # Set the variables $source, $target and the like.
     new_recdict["buildtarget"] = target.get_name()
     dl = shorten_dictlist(dep.targetlist)
@@ -105,9 +111,9 @@ def exec_commands(dep, recdict, target, filetype = None):
 
     # Use "var_ABC" attributes from dependencies to set the "ABC" variable.
     # Use "add_ABC" attributes from dependencies to add to the "ABC" variable.
-    from Action import get_var_attr
+    from Action import get_vars_from_attr
     for s in dl:
-        get_var_attr(s, new_recdict)
+        get_vars_from_attr(s, new_recdict)
 
     # Create a ParsePos object to contain the parse position in the string.
     # Make a copy of the RecPos stack, so that the item on top can be
@@ -350,7 +356,11 @@ def target_update_inlist(work, recdict, scope_recdict,
     done = build_done_NOT
     for src_name in names:
         # Find the node after removing "./" things ($SRCPATH may cause it)
-        node = work.get_node(os.path.normpath(src_name), 1, src_dict)
+        if is_url(src_name):
+            norm_name = src_name
+        else:
+            norm_name = os.path.normpath(src_name)
+        node = work.get_node(norm_name, 1, src_dict)
 
         # Check if we skip updating this item.
         if node.attributes.get("update") == "no":
@@ -411,8 +421,6 @@ def dictlist_update(dictlist, work, recdict, scope_recdict,
     else:
         nlevel = None
 
-    from Remote import is_url
-
     for src_dict in dictlist:
         # For an absolute path use the source name literally, otherwise use
         # $SRCPATH to locate the source.
@@ -457,7 +465,7 @@ def dictlist_update(dictlist, work, recdict, scope_recdict,
                 # Quit building at the first error.
                 raise UserError, msg
 
-        if autodepend and not node.attributes.get("virtual"):
+        if autodepend and node and not node.attributes.get("virtual"):
             # Find automatic dependencies for this item and update them
             # recursively.
             autodep_update(work, recdict, scope_recdict,
@@ -492,6 +500,8 @@ def check_need_update(recdict, update, src_dict, target, rootname = None):
             node = src_dict["_node"]
             src_name = node.get_name()
             virtual = node.attributes.get("virtual")
+        elif is_url(src_dict["name"]):
+            src_name = src_dict["name"]
         else:
             src_name = os.path.normpath(src_dict["name"])
 
@@ -503,7 +513,10 @@ def check_need_update(recdict, update, src_dict, target, rootname = None):
         elif virtual:
             update.source = src_name
         else:
-            check = check_name(recdict, src_name, src_dict)
+            alt_dict = None
+            if src_dict.has_key("_node"):
+                alt_dict = src_dict["_node"].attributes
+            check = check_name(recdict, src_name, src_dict, alt_dict)
             if check == "newer":
                 c = "time"              # "newer" check also uses timestamp
             else:
@@ -525,7 +538,7 @@ def check_need_update(recdict, update, src_dict, target, rootname = None):
             # If the check is "newer" need to find the newest timestamp.
             # For other checks building is to be done if the sign differs.
             if check == "newer":
-                t = long(new)
+                t = long(float(new))
                 if update.time == 0 or t > update.time:
                     update.time = t
                     update.time_source = src_name
@@ -593,15 +606,16 @@ def buildcheck_update(recdict, checkstring, dep, target, update, level):
 
     # Use "var_xxx" attributes from source files, they overrule variables.
     # Also use "var_xxx" attributes from target files.  This isn't exactly the
-    # same as what happens for an executed dependency, but it is require to
+    # same as what happens for an executed dependency, but it is required to
     # handle attributes that are passed to an action, such as
     # {var_LDFLAGS = -X}.
-    from Action import get_var_attr
+    from Action import get_vars_from_attr
+    save_attr = {}
     for s in dep.sourcelist + dep.targetlist:
         node = work.find_node(s["name"])
         if node:
-            get_var_attr(node.attributes, recdict, save)
-        get_var_attr(s, recdict, save)
+            get_vars_from_attr(node.attributes, recdict, save_attr)
+        get_vars_from_attr(s, recdict, save_attr)
 
     # Always compute the check signature, it either needs to be compared (not
     # outdate yet) or stored (when outdated already).
@@ -617,7 +631,7 @@ def buildcheck_update(recdict, checkstring, dep, target, update, level):
     if recdict.has_key("xcommands"):
         del recdict["xcommands"]
 
-    for k in saved:
+    for k in saved + save_attr.keys():
         if save[k] is None:
             if recdict.has_key(k):
                 del recdict[k]
@@ -631,7 +645,7 @@ def buildcheck_update(recdict, checkstring, dep, target, update, level):
     # have irrelevant trailing spaces.  Leading white space may be relevant
     # though, keep it.
     # TODO: how about white space inside quotes?
-    str = ''
+    bstr = ''
     i = 0
     leading = 1
     check_str_len = len(check_str)
@@ -649,11 +663,11 @@ def buildcheck_update(recdict, checkstring, dep, target, update, level):
         while (e < check_str_len and check_str[e] != ' '
                                                      and check_str[e] != '\t'):
             e = e + 1
-        str = str + check_str[s:e]
+        bstr = bstr + check_str[s:e]
         i = e
-        leading = (str[-1] == '\n')
+        leading = (bstr[-1] == '\n')
 
-    update.set_buildcheck(recdict, str, target, level)
+    update.set_buildcheck(recdict, bstr, target, level)
 
 
 def target_rule(work, recdict, target, update, src_list_list, level):
@@ -684,13 +698,15 @@ def target_rule(work, recdict, target, update, src_list_list, level):
     for r in work.rules:
         # Check if the rule exists in the scope of the target
         if r.scope == "local":
-            if not target.scope_recdict is r.buildrecdict:
+            # The rule must be defined in the same recipe as the build
+            # commands.
+            if not target.scope_recdict["_recipe"] is r.buildrecdict["_recipe"]:
                 continue
         elif r.scope == "normal":
             if not rule_in_tree(r.buildrecdict, target.scope_recdict):
                 continue
 
-        matchstr, dir, matchlen = r.match_target(target_name, target_sname)
+        matchstr, adir, matchlen = r.match_target(target_name, target_sname)
         if matchstr:
             src_list = r.target2sourcelist(target_name, target_sname)
             # Skip rules where the target and source are equal.
@@ -790,7 +806,7 @@ def target_rule(work, recdict, target, update, src_list_list, level):
                         buildrule.commands,
                         ".",
                         buildrule.buildrecdict)
-        dep.matchstr, dir, matchlen = buildrule.match_target(target_name,
+        dep.matchstr, adir, matchlen = buildrule.match_target(target_name,
                                                               target_sname)
 
         # Apply attributes from the rule to the target.  Must be before
@@ -1160,6 +1176,30 @@ def dictlist2depnodelist(work, dictlist):
     return nl
 
 
+def commands_with_sections(s):
+    """
+    Return non-zero when sections exist in command block "s".
+    A section starts with ">always", ">build" or ">nobuild".
+    Ignore white space and comments.
+    """
+    slen = len(s)
+    i = 0
+    while i < slen:
+        i = skip_white(s, i)
+        if i >= slen:
+            break
+        c = s[i]
+        if c == '>':
+            return 1
+        if c != '\n' and c != '#':
+            break
+        i = string.find(s, '\n', i)
+        if i < 0:
+            break
+        i = i + 1
+    return 0
+
+
 def may_exec_depend(recdict, dep, update, target, src_list_list, level,
                                                               filetype = None):
     """Execute the build commands of dependency "dep" if the target "target" is
@@ -1198,15 +1238,29 @@ def may_exec_depend(recdict, dep, update, target, src_list_list, level,
         # building failed previously.
         return retval
 
+    # We do the building when the target needs to be updated AND when the build
+    # commands start with ">always", ">build" or ">nobuild".
     if reason:
-        if (Global.cmd_args.options.get("touch") and not Global.force_build):
+        build = 1
+    else:
+        msg_depend(recdict, _('Target "%s" is up-to-date') % target_sname,
+                                                                         level)
+        build = commands_with_sections(dep.commands)
+
+    if build:
+        if (reason and Global.cmd_args.options.get("touch")
+                                                   and not Global.force_build):
             # Skip standard virtual targets, they are always rebuild
             if not target.name in Global.virtual_targets:
                 msg_info(recdict, _('Touching "%s"') % target_sname)
         else:
+            if reason:
+                msg = reason
+            else:
+                msg = _("command block with sections")
             msg_depend(recdict, _('Updating "%s" from "%s": %s')
                                                                % (target_sname,
-                       dictlist2str(shorten_dictlist(dep.sourcelist)), reason),
+                       dictlist2str(shorten_dictlist(dep.sourcelist)), msg),
                                                                level)
 
         # Skip when not actually building.
@@ -1217,10 +1271,11 @@ def may_exec_depend(recdict, dep, update, target, src_list_list, level,
                 aap_checkdir(dep.rpstack, recdict, target_name)
 
         # Execute the build commands.
-        exec_commands(dep, recdict, target, filetype = filetype)
+        exec_commands(dep, recdict, target, filetype = filetype,
+                                                         really_build = reason)
 
         # Skip when not actually building.
-        if not skip_commands():
+        if reason and not skip_commands():
             # Check that the target was really updated.  Only when it
             # exists.
             if dst_time > 0 and not target.attributes.get("virtual"):
@@ -1231,10 +1286,6 @@ def may_exec_depend(recdict, dep, update, target, src_list_list, level,
                 #   Could give an error here, but on fast machines it often
                 #   happens while nothing is wrong.
 
-    else:
-        msg_depend(recdict, _('Target "%s" is up-to-date') % target_sname,
-                                                                         level)
-
     if reason or update.newsig:
         # Update the signatures for all targets.  We assume they were all
         # updated, even though the build commands may skip some or were not
@@ -1531,6 +1582,7 @@ def dobuild(work, target = None):
         targets = default_targets(work, work.recdict)
 
     # Update the toplevel targets.  Catch a ":quit" or ":exit" command.
+    exitobj = None
     try:
         for target in targets:
 
@@ -1556,8 +1608,8 @@ def dobuild(work, target = None):
                     # Quit building at the first error.
                     raise UserError, msg
 
-    except NormalExit:
-        pass
+    except NormalExit, n:
+        exitobj = n
 
     # Update the "finally" target if it exists.
     target = work.find_node("finally", "")
@@ -1565,5 +1617,8 @@ def dobuild(work, target = None):
         msg_depend(work.recdict, _('Building the "finally" target'))
         target_update(work, work.recdict, target, level = 1)
 
+    if exitobj != None:
+        raise exitobj
+
 
 # vim: set sw=4 et sts=4 tw=79 fo+=l:
