@@ -21,6 +21,7 @@ import import_re        # import the re module in a special way
 import string
 import tempfile
 
+import Action
 from Dictlist import str2dictlist, list2str, dictlist2str, listitem2str
 from Dictlist import dictlist_expand, parse_attr
 from Error import *
@@ -192,6 +193,17 @@ def childdir(arg):
         return ''
     return relativedir(arg, "CHILDDIR")
 
+def parentdir(arg):
+    """
+    Return the argument expanded to be relative to the current recipe
+    directory, assuming it is relative to the parent recipe.
+    """
+    recdict = Global.globals
+    if not recdict.has_key("PARENTDIR"):
+        msg_error(recdict, _("Using parentdir() at the top level"))
+        return ''
+    return relativedir(arg, "PARENTDIR")
+
 def topdir(arg):
     """
     Return the argument expanded to be relative to the directory of the
@@ -199,6 +211,13 @@ def topdir(arg):
     """
     return relativedir(arg, "TOPDIR")
 
+def basedir(arg):
+    """
+    Return the argument expanded to be relative to the current recipe
+    directory, assuming it is relative to the toplevel recipe.
+    """
+    return relativedir(arg, "BASEDIR")
+
 def relativedir(arg, varname):
     """
     Implementation of childdir() and topdir()
@@ -318,6 +337,30 @@ def srcitem2obj(recdict, name, attrdict = {}, sufname = "OBJSUF", auto = 1):
     return n
 
 
+def bdir(name, with_attr = 1):
+    """Prepend $BDIR to name if there is no BDIR there already.
+       If (optional) with_attr is 0, strip attributes and return
+       only a filename."""
+    bdir = get_var_val(0, Global.globals, "_no","BDIR")
+    bdirprefix = "build-" + get_var_val(0, Global.globals, "_no", "OSNAME")
+
+    res = var2dictlist(name)
+    for i in res:
+        filename = i["name"]
+        # Common case: no bdir found. Prepend bdir.
+        if filename.find(bdirprefix) < 0 :
+            i["name"] = os.path.join(bdir,filename)
+
+    # Convert back to string, with or without attributes
+    if with_attr:
+        return dictlist2str(res)
+    else:
+        rv = ""
+        for i in res:
+            rv += (i["name"]+" ")
+        return rv
+
+
 def get_attr(fname):
     """Obtain the attribute dictionary for a file name.
        Only works in a recipe, because it uses Global.globals."""
@@ -394,11 +437,11 @@ def program_path(name, path = None, pathext = None, skip = None):
     if string.find(os.path.basename(name), '.') > 0:
         extlist = [ '' ]
 
-    for dir in dirlist:
-        if skip and os.path.abspath(dir) == os.path.abspath(skip):
+    for adir in dirlist:
+        if skip and os.path.abspath(adir) == os.path.abspath(skip):
             continue
         for ext in extlist:
-            fullname = os.path.join(dir, name) + ext
+            fullname = os.path.join(adir, name) + ext
             if os.path.isfile(fullname):
                 try:
                     # If os.access() is available we can check if the file is
@@ -605,6 +648,45 @@ def aap_has(arg):
     return 0
 
 
+def define_action(action, primary, commands, attr = {},
+                            intypes = [], outtypes = [],
+                            defer_var_names = [],
+                            outfilename = ''):
+    """
+    Define an action.
+    """
+    from RecPos import RecPos
+
+    # TODO: is there a way to put a line number in rpstack?
+    recdict = Global.globals
+    s = "define_action(%s)" % action
+    if intypes:
+        s = s + " from %s" % reduce(lambda x, y: x + "," + y, intypes)
+    if outtypes:
+        s = s + " to %s" % reduce(lambda x, y: x + "," + y, outtypes)
+    act = Action.Action([RecPos(s)], recdict, attr, commands,
+                                    intypes = intypes, outtypes = outtypes,
+                                             defer_var_names = defer_var_names,
+                                             outfilename = outfilename)
+    Action.action_add_to_dict(action, act)
+    act.primary = primary
+    return act
+
+
+def action_in_types(name, outtype = None):
+    """
+    Find action object(s) by name and return a list of the supported input
+    types.  Restrict to actions that support "outtype".
+    """
+    return Action.get_ftypes(name, 1, outtype = outtype)
+
+def action_out_types(name, intype = None):
+    """
+    Find action object(s) by name and return a list of the supported output
+    types.  Restrict to actions that support "intype".
+    """
+    return Action.get_ftypes(name, 0, intype = intype)
+
 # Remember the last installed port to avoid asking the user twice about using
 # it.
 lastportname = None
@@ -743,24 +825,24 @@ def ask_prefix(name):
         if r[0] == 'q' or r[0] == 'Q':
             recipe_error([], _('ask_prefix() aborted'))
         if r[0] == '2' or r[0] == '4':
-            dir = raw_input(_("Enter directory name: "))
-            while dir and (dir[-1] == '\n' or dir[-1] == '\r'):
-                dir = dir[:-1]
-            dir = os.path.expanduser(dir)
+            adir = raw_input(_("Enter directory name: "))
+            while adir and (adir[-1] == '\n' or adir[-1] == '\r'):
+                adir = adir[:-1]
+            adir = os.path.expanduser(adir)
         elif r[0] == '1':
-            dir = "/usr/local/"
+            adir = "/usr/local/"
         elif r[0] == '3':
-            dir = home
+            adir = home
         else:
             return 0, ''
 
         # Make sure the directory ends in a slash (or backslash).
-        if dir[-1] != '/' and dir[-1] != '\\':
-            dir = dir + '/'
+        if adir[-1] != '/' and adir[-1] != '\\':
+            adir = adir + '/'
 
         if r[0] == '1' or r[0] == '2':
-            return 1, dir
-        return 0, dir
+            return 1, adir
+        return 0, adir
 
     # invalid answer
     return 0, ''
@@ -843,10 +925,12 @@ def aap_depend_c(recdict, sourcestr, targetstr, local = 1, flags = None):
     # "scanned" is a list of files that have been encountered.  The key is an
     # absolute file name.  The value is zero when the file is to be scanned,
     # one if it has been scanned.
-    scanned = {os.path.abspath(source) : 0}
+    scanned = {full_fname(source) : 0}
 
     # Make the local pathlist from "-Idir" arguments.
-    localpathlist = [ "." ]
+    # Should we include the current directory?  Probably not.  We do look in
+    # the directory of the source file when double quotes are used.
+    localpathlist = [ ]
     if flags is None:
         flags = get_var_val_int(recdict, "CFLAGS")
     for n in [get_var_val_int(recdict, "CPPFLAGS"),
@@ -879,7 +963,7 @@ def aap_depend_c(recdict, sourcestr, targetstr, local = 1, flags = None):
                                                             % target) + str(e))
     else:
         try:
-            f.write("%s : " % source)
+            f.write("%s : " % fname_fold(source))
             for k in scanned.keys():
                 f.write(listitem2str(shorten_name(k)) + " ")
             f.close()
@@ -898,11 +982,11 @@ def _scan_c_file(recdict, fname, scanned, local, localpathlist, globalpathlist):
     def search_path(pathlist, fname):
         """Search for included file "fname" in list of dirs "pathlist"."""
         if os.path.isabs(fname):
-            return os.path.normpath(fname)
-        for dir in pathlist:
-            f = os.path.join(dir, fname)
+            return fname
+        for adir in pathlist:
+            f = os.path.join(adir, fname)
             if os.path.exists(f):
-                return os.path.abspath(f)
+                return full_fname(f)
         return None     # include file not found
 
     def search_target(work, pathlist, fname):
@@ -914,11 +998,11 @@ def _scan_c_file(recdict, fname, scanned, local, localpathlist, globalpathlist):
         n = work.find_node(fname)
         if n and (n.attributes.get("fetch") or n.get_dependencies()):
             return fname
-        for dir in pathlist:
-            f = os.path.join(dir, fname)
+        for adir in pathlist:
+            f = os.path.join(adir, fname)
             n = work.find_node(f)
             if n and (n.attributes.get("fetch") or n.get_dependencies()):
-                return os.path.abspath(f)
+                return full_fname(f)
         return None     # include file not found
 
     try:
@@ -933,6 +1017,10 @@ def _scan_c_file(recdict, fname, scanned, local, localpathlist, globalpathlist):
                 msg_extra(recdict, _('Cannot find included file "%s"' % fname))
                 return
 
+        mypath = os.path.dirname(fname)
+        if mypath == '':
+            mypath = '.'
+
         f = open(fname)
         msg_extra(recdict, _('Scanning "%s" for dependencies') % fname)
 
@@ -966,19 +1054,22 @@ def _scan_c_file(recdict, fname, scanned, local, localpathlist, globalpathlist):
                             while i < line_len and line[i] != quote:
                                 i = i + 1
                             if i < line_len and line[i] == quote:
+                                # Make file name uniform (forward slashes).
+                                name = fname_fold(line[s:i])
+
                                 # Search for the file in the specified path for
                                 # include files.
                                 if quote == '"':
-                                    pathlist = localpathlist + globalpathlist
+                                    pathlist = [ mypath ] + localpathlist + globalpathlist
                                 else:
                                     pathlist = globalpathlist + localpathlist
-                                fn = search_path(pathlist, line[s:i])
+                                fn = search_path(pathlist, name)
 
                                 # If the file is not found search for a target
                                 # that can produce it or fetch it.
                                 if not fn:
                                     fn = search_target(getwork(recdict),
-                                                           pathlist, line[s:i])
+                                                                pathlist, name)
                                 if fn and not scanned.has_key(fn):
                                     scanned[fn] = not local
 
@@ -1063,7 +1154,10 @@ def install_files(desttail, source, modestr, strip = 0, mandir = 0,
         elif item.get("keepdir"):
             destloc = os.path.join(destloc, os.path.dirname(name))
 
-        destname = os.path.join(destloc, os.path.basename(name))
+        if item.get("installname"):
+            destname = os.path.join(destloc, item.get("installname"))
+        else:
+            destname = os.path.join(destloc, os.path.basename(name))
         scheme, mach, path = url_split3(destname)
 
         if uninstall:
